import numpy as np


def find_nearest_idx(array, value):
    """Return the index of nearest value in an array to the given value"""
    idx = (np.abs(array-value)).argmin()
    return idx


def find_nearest(array, value):
    """Return the nearest value in an array to the given value"""
    return array[find_nearest_idx(array, value)]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import sys
sys.path.insert(0, os.path.abspath('..'))
import sphinx_rtd_theme

# -- General configuration ------------------------------------------------
autoclass_content = "both"  # include both class docstring and __init__
autodoc_default_flags = [
        # Make sure that any autodoc declarations show the right members
        "members",
        "inherited-members",
        "private-members",
        "show-inheritance",
]
autosummary_generate = True  # Make _autosummary files and include them
napoleon_numpy_docstring = False  # Force consistency, leave only Google
napoleon_use_rtype = False  # More legible

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx_automodapi.automodapi',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
numpydoc_show_class_members = False
source_suffix = '.rst'
master_doc = 'index'
project = 'python-graspfile'
year = '2019-2020'
author = 'Paul Grimes'
copyright = '{0}, {1}'.format(year, author)
version = release = '0.3.0'

pygments_style = 'trac'
templates_path = ['.']
extlinks = {
    'issue': ('https://github.com/Smithsonian/python-graspfile/issues/%s', '#'),
    'pr': ('https://github.com/Smithsonian/python-graspfile/pull/%s', 'PR #'),
}


html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_theme_options = {
}

html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
    '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



# test_file.py

import io

import graspfile.torfile
import pytest

test_file = "tests/test_data/tor_files/python-graspfile-example.tor"
"""TICRA Tools 10.0.1 GRASP .tor file"""


@pytest.fixture
def empty_tor_file():
    """Return an empty GraspTorFile instance."""
    return graspfile.torfile.GraspTorFile()


@pytest.fixture
def input_file_object():
    """Return a file object pointing to the test file."""
    return open(test_file)


@pytest.fixture
def filled_tor_file(empty_tor_file, input_file_object):
    """Return a GraspTorFile instance filled from the tor_file"""
    empty_tor_file.read(input_file_object)
    input_file_object.close()
    return empty_tor_file


def test_loading_tor_file(filled_tor_file):
    """Test loading from a tor cutfile"""
    # Check that something was loaded
    assert len(filled_tor_file.keys()) > 0

    # Check that the frequencies were loaded
    assert len(filled_tor_file["single_frequencies"].keys()) > 0


def test_reloading_tor_file(filled_tor_file):
    """Test outputting the filled_tor_file to text and reloading it with StringIO"""
    test_str = repr(filled_tor_file)

    try:
        test_io = io.StringIO(test_str)
    except TypeError:
        test_io = io.StringIO(unicode(test_str))

    reload_tor_file = graspfile.torfile.GraspTorFile(test_io)

    assert len(filled_tor_file.keys()) == len(reload_tor_file.keys())



"""A class to hold a parsed GRASP Tor File in a collection of objects"""

from collections import OrderedDict

import graspfile.torparser as torparser

_debug_ = False

# trick for py2/3 compatibility
if 'basestring' not in globals():
    basestring = str

"""List of acceptable GraspTorObject types"""
grasp_object_types = [""]


class GraspTorValue:
    """A container for values from GraspTorMember objects"""

    def __init__(self, tor_value="_None"):
        """Container for values within GraspTorMember, GraspTorStruct, and GraspTorSequence objects.

        Args:
            tor_value: a `pyparsing.ParseResults` object from a tor_value tokentcOK.
        """
        #: str: The value as a string
        self.value = None

        #: str: The unit of the value as a string
        self.unit = None

        if tor_value != "_None":
            self.fill(tor_value)

    def __repr__(self):
        """Return a useful string representation of the GraspTorValue object.

        Returns:
            string representation of GraspTorValue in format suitable for use in a .tor file."""
        if self.unit:
            return repr(self.value) + " " + self.unit
        else:
            if isinstance(self.value, basestring):
                return self.value
            else:
                return repr(self.value)

    def fill(self, tor_value):
        """Fills the GraspTorValue object from output by the parser.

        Args:
            tor_value: a `pyparsing.ParseResults` class output by the torparser module.
        """
        if _debug_:
            print("GraspTorValue.fill received: {:}".format(tor_value))

        try:
            if isinstance(tor_value, basestring):
                self.value = tor_value
            else:
                if len(tor_value) > 1:
                    self.value = tor_value[0]
                    self.unit = tor_value[1]
                else:
                    self.value = tor_value[0]
        except TypeError:
            if _debug_:
                print("TypeError caught for tor_value = {:}".format(tor_value))
            self.value = tor_value


class GraspTorMember:
    """A container for the member parameter of an GraspTorObject """

    def __init__(self, tor_member=None):
        self.name = None

        self._type = None

        self._value = None
        if tor_member:
            self.fill(tor_member)

    def __repr__(self):
        """Return a useful string representation of the GraspTorMember object."""
        return self._value.__repr__()

    def fill(self, tor_member):
        if _debug_:
            print("GraspTorMember.fill received: {:}".format(tor_member))

        self.name = tor_member[0]
        if len(tor_member) > 2:
            self._type = tor_member[1]
        else:
            self._type = "value"

        if self._type == "struct":
            self._value = GraspTorStruct(tor_member[1:])
        elif self._type == "ref":
            self._value = GraspTorRef(tor_member[1:])
        elif self._type == "sequence":
            self._value = GraspTorSequence(tor_member[1:])
        else:
            self._value = GraspTorValue(tor_member[1])

    @property
    def type(self):
        """Return the type of the GraspTorMember"""
        return self._type

    @property
    def value(self):
        """Short circuit through tor_value to supply tV.value if it is a simple value"""
        if self._type == "value":
            return self._value.value
        else:
            return self._value

    @value.setter
    def value(self, new_val):
        if self._type == "value":
            if isinstance(new_val, GraspTorValue):
                self._value = new_val
            else:
                self._value.value = new_val
        else:
            self._value = new_val

    @property
    def unit(self):
        """Short circuit through tor_value to supply tV.unit if appropriate, else return None"""
        if self._type == "value":
            return self._value.unit
        else:
            return None

    @unit.setter
    def unit(self, new_unit):
        if self._type == "value":
            self._value.unit = new_unit
        else:
            pass

    def __getitem__(self, key):
        return self.value[key].value

    def __setitem__(self, key, new_value):
        self.value[key].value = new_value


class GraspTorRef:
    """A container for a value that is a reference to another GraspTorObject"""

    def __init__(self, tor_ref=None):

        #: str: Reference to another GraspTorObject
        self.ref = None
        if tor_ref:
            self.fill(tor_ref)

    def __repr__(self):
        """Return a useful string representation of the GraspTorRef object."""
        return "ref({:s})".format(self.ref)

    def fill(self, tor_ref):
        if _debug_:
            print("GraspTorRef.fill received: {:}".format(tor_ref))
        self.ref = tor_ref[1]


class GraspTorSequence(list):
    """A container for a value that is a sequence of GraspTorValues."""

    def __init__(self, tor_seq=None):
        super(GraspTorSequence, self).__init__()
        if tor_seq:
            self.fill(tor_seq)

    def __repr__(self):
        """Return a useful string representation of the GraspTorSequence object."""
        outstring = "sequence("
        for v in self:
            outstring += (repr(v)) + ","
        outstring = outstring[:-1] + ")"

        return outstring

    def fill(self, tor_seq):
        if _debug_:
            print("GraspTorSequence.fill received: {:}".format(tor_seq))

        for t in tor_seq[1:]:
            self.append(GraspTorValue(t))


class GraspTorStruct(OrderedDict):
    """A container for a GraspTorStruct, that has a number of members.  Members are
    stored as an OrderedDict."""

    def __init__(self, tor_struct=None):
        OrderedDict.__init__(self)
        if tor_struct:
            self.fill(tor_struct)
        else:
            pass

    def __repr__(self):
        """Return a useful string representation of the GraspTorSequence object."""
        outstring = "struct("
        for v in iter(self.keys()):
            outstring += v + ": " + repr(self[v]) + ", "
        outstring = outstring[:-2] + ")"

        return outstring

    def fill(self, tor_struct):
        """Fill the GraspTorObject using the pyparsing results"""
        if _debug_:
            print("GraspTorStruct.fill received: {:}".format(tor_struct))

        for t in tor_struct[1:]:
            self[t[0]] = GraspTorMember(t)


class GraspTorComment:
    """A container for comments from a GraspTorFile"""

    def __init__(self, tor_comment=None):
        #: str: Name of comment object
        self.name = None

        #: str: The type of the object
        self._type = "comment"

        #: str: Test of the comment object
        self.text = None

        #: int: Line number that the comment appears in.
        self.location = None
        if tor_comment:
            self.fill(tor_comment)

    def __repr__(self):
        """Return the comment as // prefixed lines"""
        return "\n".join(self.text)

    def fill(self, tor_comment):
        if _debug_:
            print("GraspTorComment.fill received: {:}".format(tor_comment))
            print("with name: {:}".format(tor_comment[0]))
        self.name = tor_comment[0]
        self.location = int(self.name.lstrip("comment"))
        self.text = tor_comment[2]

    @property
    def type(self):
        """Return the type of the comment object as "commment"."""
        return self._type


class GraspTorObject(OrderedDict):
    """A container for a GraspTorObject, that has a name, a type and a number of members.  Members are
    stored as an OrderedDict."""

    def __init__(self, tor_obj=None):
        OrderedDict.__init__(self)
        self._name = None
        self._type = None

        if isinstance(tor_obj, str):
            self.read_str(tor_obj)
        elif isinstance(tor_obj, torparser.pp.ParseResults):
            self.fill(tor_obj)
        else:
            pass

    def __repr__(self):
        """Return a useful string representation of the GraspTorObject object."""
        if self.type == "comment":
            outstring = repr(self["comment"]) + "\n"
        else:
            memberstrings = []
            for k in iter(self.keys()):
                memberstrings.append(k + "   : " + repr(self[k]))

            memberstring = ",\n  ".join(memberstrings)

            outstring = """{:}  {:}
(
  {:}
)
    """.format(self._name, self._type, memberstring)

        return outstring

    def read_str(self, tor_str):
        """Read the contents of the string into a tor_object and then process the results"""
        res = torparser.tor_object.parseString(tor_str)
        self.fill(res)

    def fill(self, tor_obj):
        """Fill the GraspTorObject using the pyparsing results"""
        if _debug_:
            print("GraspTorObject.fill received: {:}".format(tor_obj))
            print("Type:", type(tor_obj))

        self._name = tor_obj[0]
        self._type = tor_obj[1]
        for r in tor_obj[2:]:
            if self._type == "comment":
                self["comment"] = GraspTorComment(tor_obj)
            else:
                self[r[0]] = GraspTorMember(r)

    @property
    def name(self):
        """Return the name of the GraspTorObject"""
        return self._name

    @name.setter
    def name(self, new_name):
        """Set the name of the GraspTorObject"""
        self._name = new_name

    @property
    def type(self):
        """Return the type of the GraspTorObject"""
        return self._type

    @type.setter
    def type(self, new_type):
        """Set the type of the GraspTorObject. Checks that type is sane"""
        # if new_type in grasp_object_types:
        #     self._type = new_type
        # else:
        #     raise ValueError("Unknown type for a GraspTorObject")
        self._type = new_type


class GraspTorFile(OrderedDict):
    """A container for objects read from a tor file.  Subclasses OrderedDict to provide a dict of torObjects
     keyed by name, and sorted by insertion order"""

    def __init__(self, file_like=None):
        """Create a TorFile object, and if fileLike is specified, read the file"""
        OrderedDict.__init__(self)
        self._parser = torparser.tor_file
        if file_like:
            self.read(file_like)

    def __repr__(self):
        """Return a GRASP readable string for the GraspTorFile object"""
        outstring = ""
        for k in iter(self.keys()):
            outstring += repr(self[k]) + "\n"

        return outstring

    def read(self, file_like):
        """Read a list of torObjects and torComments from a fileLike object.  We use pyparsing.ParserElement's parseFile
        method, which can take either a file-like object or a filename to open.  If you wish to parse an existing string
        object, used StringIO to supply a file-like object containing the string."""
        # Parse the file
        res = self._parser.parseFile(file_like)

        # Turn the parse results into objects
        self.fill(res)

    def fill(self, tor_file):
        """Fill the GraspTorFile using the parser results in tor_file"""
        for r in tor_file:
            temp = GraspTorObject(r)
            self[temp.name] = temp



# test_analysis_grid.py

import numpy as np
import pytest
from graspfile import grid
from graspfile.analysis import grid as ga

test_grid_file = "tests/test_data/grasp_files/square_aperture.grd"
"""TICRA Tools 19.1 GRASP Grid file, consisting of three grids at 82, 97 and 112 GHz."""


@pytest.fixture
def empty_grasp_grid():
    """Return an empty GraspGrid instance."""
    return grid.GraspGrid()


@pytest.fixture
def grid_file():
    """Return a file object that points to a GRASP Grid file"""
    return open(test_grid_file)


@pytest.fixture
def filled_grasp_grid(empty_grasp_grid, grid_file):
    """Return a GraspGrid instance filled from the grid_file."""
    empty_grasp_grid.read(grid_file)
    grid_file.close()
    return empty_grasp_grid


@pytest.fixture
def filled_grasp_field(filled_grasp_grid):
    """Return a GraspField instance from the filled_grasp_grid fixture"""
    return filled_grasp_grid.fields[0]


def test_get_value(filled_grasp_field):
    """Test getting a value from the field"""
    x_vals, y_vals = filled_grasp_field.positions_1d

    # Test values of xv, yv in x_vals, y_vals
    xv = x_vals[int(len(x_vals)/3)]
    yv = y_vals[int(2*len(y_vals)/3)]

    value = filled_grasp_field.get_value(xv, yv)

    assert len(value) == filled_grasp_field.field_components

    # Test values of xv, yv not in x_vals, y_vals
    xv = (x_vals[int(len(x_vals)/3)] + x_vals[int(len(x_vals)/3)+1])/2
    yv = (y_vals[int(2*len(y_vals)/3)] + y_vals[int(2*len(y_vals)/3)-1])/2

    value = filled_grasp_field.get_value(xv, yv)

    assert len(value) == filled_grasp_field.field_components


def test_finding_peak(filled_grasp_field):
    """Test find the peak of a grasp field"""
    x_peak, y_peak = ga.find_peak(filled_grasp_field)

    assert x_peak >= filled_grasp_field.grid_min_x
    assert x_peak <= filled_grasp_field.grid_max_x
    assert y_peak >= filled_grasp_field.grid_min_y
    assert y_peak <= filled_grasp_field.grid_max_y


def test_finding_peak_comp1(filled_grasp_field):
    """Test find the peak of a grasp field"""
    x_peak, y_peak = ga.find_peak(filled_grasp_field, comp=1)

    assert x_peak >= filled_grasp_field.grid_min_x
    assert x_peak <= filled_grasp_field.grid_max_x
    assert y_peak >= filled_grasp_field.grid_min_y
    assert y_peak <= filled_grasp_field.grid_max_y


def test_finding_peak_min_radius(filled_grasp_field):
    """Test find the peak of a grasp field"""
    radii = filled_grasp_field.radius_grid()

    shape = radii.shape
    min_radius = radii[int(shape[0]/3), int(shape[1]/3)]

    x_peak, y_peak = ga.find_peak(filled_grasp_field, min_radius=min_radius)

    assert x_peak >= filled_grasp_field.grid_min_x
    assert x_peak <= filled_grasp_field.grid_max_x
    assert y_peak >= filled_grasp_field.grid_min_y
    assert y_peak <= filled_grasp_field.grid_max_y


def test_finding_peak_max_radius(filled_grasp_field):
    """Test find the peak of a grasp field"""
    radii = filled_grasp_field.radius_grid()

    shape = radii.shape
    max_radius = radii[int(shape[0] / 5), int(shape[1] / 5)]

    x_peak, y_peak = ga.find_peak(filled_grasp_field, max_radius=max_radius)

    assert x_peak >= filled_grasp_field.grid_min_x
    assert x_peak <= filled_grasp_field.grid_max_x
    assert y_peak >= filled_grasp_field.grid_min_y
    assert y_peak <= filled_grasp_field.grid_max_y


def test_finding_center(filled_grasp_field):
    """Test find the center of a grasp field"""
    x_cent, y_cent = ga.find_center(filled_grasp_field)

    assert x_cent >= filled_grasp_field.grid_min_x
    assert x_cent <= filled_grasp_field.grid_max_x
    assert y_cent >= filled_grasp_field.grid_min_y
    assert y_cent <= filled_grasp_field.grid_max_y


def test_finding_center_comp1(filled_grasp_field):
    """Test find the center of a grasp field"""
    x_cent, y_cent = ga.find_center(filled_grasp_field, comp=1)

    assert x_cent >= filled_grasp_field.grid_min_x
    assert x_cent <= filled_grasp_field.grid_max_x
    assert y_cent >= filled_grasp_field.grid_min_y
    assert y_cent <= filled_grasp_field.grid_max_y


def test_finding_center_trunc(filled_grasp_field):
    """Test find the center of a grasp field"""
    x_peak, y_peak = ga.find_peak(filled_grasp_field)
    peak = filled_grasp_field.get_value(x_peak, y_peak)

    trunc = np.abs(peak[0])/5.0

    x_cent, y_cent = ga.find_center(filled_grasp_field, trunc_level=trunc)

    assert x_cent >= filled_grasp_field.grid_min_x
    assert x_cent <= filled_grasp_field.grid_max_x
    assert y_cent >= filled_grasp_field.grid_min_y
    assert y_cent <= filled_grasp_field.grid_max_y


def test_finding_center_min_radius(filled_grasp_field):
    """Test find the center of a grasp field"""
    radii = filled_grasp_field.radius_grid()

    shape = radii.shape
    min_radius = radii[int(shape[0]/3), int(shape[1]/3)]

    x_cent, y_cent = ga.find_center(filled_grasp_field, min_radius=min_radius)

    assert x_cent >= filled_grasp_field.grid_min_x
    assert x_cent <= filled_grasp_field.grid_max_x
    assert y_cent >= filled_grasp_field.grid_min_y
    assert y_cent <= filled_grasp_field.grid_max_y


def test_finding_center_max_radius(filled_grasp_field):
    """Test find the center of a grasp field"""
    radii = filled_grasp_field.radius_grid()

    shape = radii.shape
    max_radius = radii[int(shape[0] / 5), int(shape[1] / 5)]

    x_cent, y_cent = ga.find_center(filled_grasp_field, max_radius=max_radius)

    assert x_cent >= filled_grasp_field.grid_min_x
    assert x_cent <= filled_grasp_field.grid_max_x
    assert y_cent >= filled_grasp_field.grid_min_y
    assert y_cent <= filled_grasp_field.grid_max_y



"""This is the module for manipulating grid files containing one or more field cuts from TICRA Tools, GRASP and CHAMP
"""

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import graspfile.numpy_utilities as numpy_utilities
import numpy


class GraspField:
    """Object holding a single dataset from a Grasp field on grid output file (``*.grd``)

    The field is held in a complex numpy array of shape ``(grid_n_x, grid_n_y, field_components)``
    where ``grid_n_x`` and ``grid_n_y`` set the number of points in the grid and ``field_components`` is the
    number of field components"""

    # This layout of array should mean that the polarisation components for a point
    # are contiguous memory, allowing rapid calculation of stokes parameters, etc.

    def __init__(self):
        # initialize storage variables
        self.beam_center = [0.0, 0.0]
        """list: Beam center in [x,y] form"""

        # Grid parameters
        self.grid_min_x = 0.0
        """float: Minimum extent of grid in 1st dimension"""

        self.grid_min_y = 0.0
        """float: Minimum extent of grid in 2nd dimension"""

        self.grid_max_x = 0.0
        """float: Maximum extent of grid in 1st dimension"""

        self.grid_max_y = 0.0
        """float: Maximum extent of grid in 2nd dimension"""

        self.grid_n_x = 0
        """int: Number of points in grid in 1st dimension"""

        self.grid_n_y = 0
        """int: Number of points in grid in 2nd dimension"""

        self.grid_step_x = 0.0
        """float: step size of grid in 1st dimension"""

        self.grid_step_y = 0.0
        """float: step size of grid in 2nd dimension"""

        self.k_limit = 0  # Is grid sparse (0=filled, 1=sparse)
        """int: defines whether grid is filled or sparse
            * 0: filled
            * 1: sparse"""

        self.field_components = 0
        """Number of field components in grid.

            Equivalent to the GRASP Grid file's ``ncomp``.
            ``2`` for far fields, ``3`` for near fields."""

        self.field = None
        """numpy.ndarray: the array of complex field components.
            the field object is numpy array of shape ``(grid_n_x, grid_n_y, field_components)``"""

    def read(self, f, field_components):
        """Reads the Grasp dataset from the file object passed in.  This assumes that
        it is being called from the graspGrid classes read(f) method, so that
        the file object is already at the start of the record"""

        # find the grid physical extents
        line = f.readline().split()
        self.grid_min_x = float(line[0])
        self.grid_min_y = float(line[1])
        self.grid_max_x = float(line[2])
        self.grid_max_y = float(line[3])
        self.field_components = field_components

        # find the number of points in the grid
        line = f.readline().split()
        self.grid_n_x = int(line[0])
        self.grid_n_y = int(line[1])
        self.k_limit = int(line[2])

        self.grid_step_x = (self.grid_max_x - self.grid_min_x) / (self.grid_n_x - 1)
        self.grid_step_y = (self.grid_max_y - self.grid_min_y) / (self.grid_n_y - 1)

        # We can now initialise the numpy arrays to hold the field data
        self.field = numpy.zeros(shape=(self.grid_n_x, self.grid_n_y, self.field_components), dtype=numpy.complex)

        for j in range(self.grid_n_y):
            # If k_limit is 1 then rows of grid are sparse (i.e. limited length)
            # read the limits from each line before reading in data
            if self.k_limit == 1:
                line = f.readline().split()
                i_s = int(line[0]) - 1
                i_e = i_s + int(line[1])
            else:
                i_s = 0
                i_e = self.grid_n_x

            # Read in the data
            for i in range(i_s, i_e):
                line = f.readline().split()
                if self.field_components == 2:
                    self.field[j, i, 0] = complex(float(line[0]), float(line[1]))
                    self.field[j, i, 1] = complex(float(line[2]), float(line[3]))
                else:
                    self.field[j, i, 0] = complex(float(line[0]), float(line[1]))
                    self.field[j, i, 1] = complex(float(line[2]), float(line[3]))
                    self.field[j, i, 2] = complex(float(line[4]), float(line[5]))

        return 0

    def index_radial_dist(self, i, j):
        """Return radial distance from the beam center to an element of the field.

        Useful for calculating the integrated power in a beam within a certain radius."""
        pos_x = self.grid_min_x + self.grid_step_x * i
        pos_y = self.grid_min_y + self.grid_step_y * j

        off_x = pos_x - self.beam_center[0]
        off_y = pos_y - self.beam_center[1]

        return numpy.sqrt(off_x ** 2 + off_y ** 2)

    @property
    def positions(self):
        """Return meshed grids of the x and y positions of each point in the field"""
        x_positions, y_positions = self.positions_1d
        return numpy.meshgrid(x_positions, y_positions)

    @property
    def positions_1d(self):
        """Return numpy arrays of the x and y positions used in the field"""
        return (numpy.linspace(self.grid_min_x, self.grid_max_x, self.grid_n_x),
                numpy.linspace(self.grid_min_y, self.grid_max_y, self.grid_n_y))

    def get_value(self, xv, yv):
        """Return the value of the field at the nearest point to xv, yv

        Args:
            xv: float containing the x coordinate of the point to get.
            yv: float containing the y coordinate of the point to get.
        Returns:
            ndarray: containing self.field_components values of the field at xv, yv"""
        x_vals, y_vals = self.positions_1d

        nx = numpy_utilities.find_nearest_idx(x_vals, xv)
        ny = numpy_utilities.find_nearest_idx(y_vals, yv)

        return self.field[nx, ny, :]

    def radius_grid(self, center=None):
        """Return an array holding the radii of each point from the beam centre.

        Args:
            center: tuple holding coordinates of the center to calculate the radius from

        Returns:
            numpy.ndarray: numpy array with same shape as the field grid holding the radii.
        """
        grid_x, grid_y = self.positions

        if center is None:
            center = self.beam_center

        return numpy.sqrt((grid_x - center[0]) ** 2 + (grid_y - center[1]) ** 2)

    def rotate_polarization(self, angle=45.0):
        """Rotate the basis of the polarization by <angle>. Will only work on linear polarization types

        ***TODO: Implement checks for polarization type***

        Args:
            angle: angle in degrees to rotate the polarization basis by.

        Returns:
            nothing.
        """
        ang = numpy.deg2rad(angle)
        output0 = self.field[:, :, 0] * numpy.cos(ang) - self.field[:, :, 1] * numpy.sin(ang)
        output1 = self.field[:, :, 1] * numpy.cos(ang) + self.field[:, :, 0] * numpy.sin(ang)
        self.field[:, :, 0] = output0
        self.field[:, :, 1] = output1


# The main file object class
class GraspGrid:
    """Object holding the data in contained in a general Grasp grid field output"""

    def __init__(self):
        """Create empty variables or lists of attributes for holding data for each dataset"""
        # Text Header
        self.header = []
        """list of str: List of lines in the header section of the file"""

        # File Type parameters
        self.ktype = 1
        """int: type of file format.

        Always ``1`` for TICRA Tools files."""

        self.nset = 0
        """int: number of grids in file."""

        self.polarization = 0
        """int: type of field components.

        Equivalent to the GRASP Grid file's ``icomp``.

        Values signify the following polarization definitions
            * `1`: Linear E_theta and E_phi.
            * `2`: Right hand and left hand circular (Erhc and Elhc).
            * `3`: Linear Eco and Ecx (Ludwig's third definition).
            * `4`: Linear along major and minor axes of the polarisation ellipse, Emaj and Emin.
            * `5`: XPD fields: E_theta/E_phi and E_phi/E_theta.
            * `6`: XPD fields: Erhc/Elhc and Elhc/Erhc.
            * `7`: XPD fields: Eco/Ecx and Ecx/Eco.
            * `8`: XPD fields: Emaj/Emin and Emin/Emaj.
            * `9`: Total power \\|E\\| and Erhc=Elhc."""

        self.field_components = 0
        """int: number of field components.

        Equivalent to the GRASP Grid file's ``ncomp``.
        ``2`` for far fields, ``3`` for near fields."""

        self.igrid = 0
        """int: grid type.

        Type of field grid.
        * `1` : uv-grid: (X; Y ) = (u; v) where u and v are the two first coordinates of the unit vector to the field
                point. Hence, r^ = u; v; p = √(1 − u² − v²) where u and v are related to the spherical angles by
                u = sin θ cos φ; v = sin θ sin φ.
        * `4` : Elevation over azimuth: (X; Y )=(Az,El), where Az and El define the direction to the field point by
                r^ = − sin Az cos El; sin El; cos Az cos El.
        * `5` : Elevation and azimuth: (X; Y )=(Az,El), where Az and El define the direction to the field point through
                the relations Az = -θ cos φ; El = θ sin φ to the spherical angles θ and φ.
        * `6` : Azimuth over elevation: (X; Y )= (Az, El), where Az and El define the direction to the field point by
                r^ = − sin Az; cos Az sin El; cos Az cos El.
        * `7` : θφ-grid: (X; Y ) = (φ; θ), where θ and φ are the spherical angles of the direction to the field point.
        * `9` : Azimuth over elevation, EDX definition: (X; Y )=(Az,El), where Az and El define the direction to the
                field point by r^ = sin Az cos El; sin El; cos Az cos El.
        * `10` : Elevation over azimuth, EDX definition: (X; Y )= (Az, El), where Az and El define the direction to the
                field point by r^ = sin Az; cos Az sin El; cos Az cos El.
        """

        self.freqs = []
        """list: List of frequencies in units of ``freq_unit``"""

        self.freq_unit = ""
        """str: The unit that the frequencies are given in."""

        self.fields = []
        """list of (:obj:`GraspField`): List of individual fields in file,"""

        self.beam_centers = []
        """list: list of beam centers for individual fields in the file."""

    def parse_header(self, header):
        """Parse the header of a grid file for useful information

        Args:
            header: list of lines in the header text.
        """
        # Use configparser to interpret the header info.
        # TO-DO:
        #   This is very dodgy, as it ignores the possibility of different frequency sets for different
        #   sources in the file, and erase the first source's information
        #   We should build a real parser for this that can handle multiple copies of keys
        config = configparser.ConfigParser(strict=False, allow_no_value=True)
        config.read_string(u"[Default]\n" + "\n".join(header))
        config = dict(config["Default"])
        # Parse the header to get the frequency information
        if "frequency" in config.keys():
            # This works for TICRA GRASP version before TICRA Tools
            res = config["frequency"]
            first, arg, rest = res.partition(":")
            if first.strip() == "start_frequency":
                # print rest
                # We have a frequency range
                start, stop, num_freq = rest.rsplit(",")
                self.freqs = numpy.linspace(float(start.split()[0]), float(stop.split()[0]), int(num_freq))
            else:
                # We probably have a list of frequencies
                # print res
                freq_strs = res.rsplit("'")
                freqs = []
                for f in freq_strs:
                    freqs.append(float(f.split()[0]))
                self.freqs = numpy.array(freqs)
        else:
            search_key = "frequencies"
            term = [key for key, val in config.items() if search_key in key][0]
            value = config[term]

            # This works for TICRA Tools versions > 19.0
            #
            # If the frequency list is long, it may spread over more than one line
            self.freq_unit = term.strip().split()[1].strip("[]")

            freq_str_list = value.split()
            freqs = []
            for f in freq_str_list:
                freqs.append(float(f))
            self.freqs = numpy.array(freqs)

    def read(self, fi):
        """Reads GRASP output grid files from file object and fills a number of variables
        and numpy arrays with the data"""

        # Loop over initial lines before "++++" getting text
        self.header = []

        while 1:
            line = fi.readline()
            if line[0:4] == "++++":
                break
            else:
                self.header.append(line)

        self.parse_header(self.header)

        # We've now got through the header text and are ready to read the general
        # field type parameters
        self.ktype = int(fi.readline())
        line = fi.readline().split()
        self.nset = int(line[0])
        self.polarization = int(line[1])
        self.field_components = int(line[2])
        self.igrid = int(line[3])

        self.beam_centers = []
        for i in range(self.nset):
            line = fi.readline().split()
            self.beam_centers.append([int(line[0]), int(line[1])])

        # field type parameters are now understood
        # we now start reading the individual fields
        for i in range(self.nset):
            dataset = GraspField()
            dataset.beamCenter = self.beam_centers[i]
            dataset.read(fi, self.field_components)
            self.fields.append(dataset)

    def rotate_polarization(self, angle=45.0):
        """Rotate the polarization basis for each field in the GraspGrid"""
        for f in self.fields:
            f.rotate_polarization(angle)



"""This is main graspfile module for manipulating files from TICRA Tools, GRASP and CHAMP

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    cut
    grid
"""

__version__ = '0.3.0'

__all__ = ['cut', 'grid', 'torfile', 'torparser']



"""pyparsing parser for GRASP .tor files"""

import pyparsing as pp

LPAREN, RPAREN, COLON, COMMA = map(pp.Suppress, "():,")

identifier = pp.Word(pp.alphas, pp.alphanums + "_")
tor_comment = pp.OneOrMore(pp.dblSlashComment)


def comment_handler(input_string, locn, tokens):
    tokenstr = "\n".join(tokens)
    mod_string = "comment{:d} comment\n(\n{:s}\n)".format(locn, tokenstr)
    mod_def = pp.Dict(pp.Group(identifier.setResultsName("_name") + identifier.setResultsName("_type") + LPAREN + pp.Group(
        pp.OneOrMore(pp.dblSlashComment)).setResultsName("text") + RPAREN))
    return mod_def.parseString(mod_string)


tor_comment.setParseAction(comment_handler)
tor_comment.setResultsName("_name")
tor_comment.setResultsName("_type")

tor_string = pp.dblQuotedString() | pp.Word(pp.alphas, pp.alphanums + "_-.")
number = pp.pyparsing_common.number()

tor_members = pp.Forward()
tor_value = pp.Forward()

tor_struct = pp.Literal("struct").setResultsName("_type") + LPAREN + pp.Dict(tor_members) + RPAREN
tor_sequence = pp.Literal("sequence").setResultsName("_type") + LPAREN + pp.delimitedList(tor_value) + RPAREN
tor_ref = pp.Literal("ref").setResultsName("_type") + LPAREN + identifier + RPAREN
tor_value << (tor_sequence | tor_ref | tor_struct | tor_string | pp.Group(number + identifier) | number)

member_def = pp.Dict(pp.Group(identifier + COLON + tor_value))
tor_members << pp.delimitedList(member_def)

object_def = pp.Group(identifier.setResultsName("_name") + identifier.setResultsName("_type") + pp.Dict(
    LPAREN + pp.Optional(tor_members) + RPAREN))
tor_object = pp.Dict(object_def | tor_comment)
tor_file = pp.Dict(pp.OneOrMore(tor_object)) + pp.stringEnd



# test_grid.py

import pytest
from graspfile import grid
from pytest import approx

test_grid_file = "tests/test_data/grasp_files/square_aperture.grd"
"""TICRA Tools 19.1 GRASP Grid file, consisting of three grids at 82, 97 and 112 GHz."""


@pytest.fixture
def empty_grasp_grid():
    """Return an empty GraspGrid instance."""
    return grid.GraspGrid()


@pytest.fixture
def grid_file():
    """Return a file object that points to a GRASP Grid file"""
    return open(test_grid_file)


@pytest.fixture
def filled_grasp_grid(empty_grasp_grid, grid_file):
    """Return a GraspGrid instance filled from the grid_file."""
    empty_grasp_grid.read(grid_file)
    grid_file.close()
    return empty_grasp_grid


@pytest.fixture
def filled_grasp_field(filled_grasp_grid):
    """Return a GraspField instance from the filled_grasp_grid fixture"""
    return filled_grasp_grid.fields[0]


def test_loading_grid(filled_grasp_grid):
    """Test loading grid from a TICRA Grid file."""
    # check that enough freqs and fields were read
    assert len(filled_grasp_grid.freqs) > 0
    assert len(filled_grasp_grid.fields) > 0

    # Check that parameters were read correctly
    assert filled_grasp_grid.ktype in [1]
    assert type(filled_grasp_grid.nset) is int
    assert filled_grasp_grid.polarization in range(1, 12)
    assert filled_grasp_grid.field_components in [2, 3]
    assert filled_grasp_grid.igrid in [2, 3, 8]

    # Check that beam centers were read correctly
    assert len(filled_grasp_grid.beam_centers) > 0
    for bc in filled_grasp_grid.beam_centers:
        assert len(bc) == 2


def test_rotate_grid_polarization(filled_grasp_grid):
    """Check that rotate_polarization runs on all fields"""
    filled_grasp_grid.rotate_polarization()
    filled_grasp_grid.rotate_polarization(angle=-45.0)


def test_loading_field(filled_grasp_field):
    """Test the individual field loaded as part of filled_grasp_grid"""
    # check that field parameters were filled correctly
    assert type(filled_grasp_field.beam_center[0]) is float
    assert type(filled_grasp_field.beam_center[1]) is float
    assert len(filled_grasp_field.beam_center) == 2

    assert type(filled_grasp_field.grid_min_x) is float
    assert type(filled_grasp_field.grid_min_y) is float
    assert type(filled_grasp_field.grid_max_x) is float
    assert type(filled_grasp_field.grid_max_y) is float
    assert type(filled_grasp_field.grid_n_x) is int
    assert type(filled_grasp_field.grid_n_y) is int
    assert type(filled_grasp_field.grid_step_x) is float
    assert type(filled_grasp_field.grid_step_y) is float

    # Check that the step values are consistent with the other grid parameters
    # (should find and use the "approx equal" test)
    assert filled_grasp_field.grid_step_x == approx(
        (filled_grasp_field.grid_max_x - filled_grasp_field.grid_min_x) / (filled_grasp_field.grid_n_x - 1))
    assert filled_grasp_field.grid_step_y == approx((filled_grasp_field.grid_max_y - filled_grasp_field.grid_min_y) / (
        filled_grasp_field.grid_n_y - 1))

    assert filled_grasp_field.k_limit in [0, 1]
    assert filled_grasp_field.field_components in [2, 3]

    # Check that the shape of the field is consistent with grid parameters
    field_shape = filled_grasp_field.field.shape

    assert field_shape[0] == filled_grasp_field.grid_n_x
    assert field_shape[1] == filled_grasp_field.grid_n_y
    assert field_shape[2] == filled_grasp_field.field_components


def test_index_radial_dist(filled_grasp_field):
    """Test the return of an array of radial distances of grid points"""
    rdist = filled_grasp_field.index_radial_dist(3, 2)
    assert rdist >= 0.0


def test_grid_pos(filled_grasp_field):
    """Test the return of the meshed grid of positions"""
    xgrid, ygrid = filled_grasp_field.positions

    assert xgrid.shape == (filled_grasp_field.grid_n_x, filled_grasp_field.grid_n_y)
    assert ygrid.shape == (filled_grasp_field.grid_n_x, filled_grasp_field.grid_n_y)


def test_radius_grid(filled_grasp_field):
    rgrid = filled_grasp_field.radius_grid()

    assert rgrid.shape == (filled_grasp_field.grid_n_x, filled_grasp_field.grid_n_y)

    rgrid2 = filled_grasp_field.radius_grid((0.1, 0.1))

    assert rgrid2.shape == (filled_grasp_field.grid_n_x, filled_grasp_field.grid_n_y)


def test_rotate_polarization(filled_grasp_field):
    ang = 180.0

    rot_field = filled_grasp_field

    rot_field.rotate_polarization(ang)
    rot_field.rotate_polarization(ang)

    assert rot_field.field == approx(filled_grasp_field.field)



"""This is the module for manipulating cut files containing one or more field cuts from TICRA Tools, GRASP and CHAMP
"""

import math

import numpy


class GraspSingleCut:
    """Class for reading, holding, manipulating and writing GRASP 9.3 format
    output cuts."""

    def __init__(self):
        self.text = ""
        """str: Cut description."""

        self.cut_type = 1
        """int: Cut type.
            Describes the type of cut.
            No GRASP equivalent

            Values signify:
                * `1`: Spherical cut
                * `2`: Planar or surface cut
                * `3`: Cylindrical cut"""

        self.v_ini = 0
        """float: Initial value of variable that cut is swept over.
            GRASP Parameter V_INI"""

        self.v_inc = 1
        """float: Increment of variable that cut is swept over.
            GRASP Parameter V_INI"""

        self.v_num = 1
        """int: Number of steps for variable that cut is swept over.
            GRASP Parameter V_NUM"""

        self.constant = 0.0
        """float: Constant variable for cut.
            GRASP Parameter C

            definition determined by ICUT."""

        self.icut = 1
        """int: Cut definition.
            GRASP Parameter ICUT

            Values signify the following cut definitions:
                For spherical cuts:
                    * `1`: Standard polar cut with fixed phi
                    * `2`: Conical cut with fixed theta

                For planar and surface cut
                    * `1`: Radial cut with fixed phi and variable distance rho
                    * `2`: Circular cut at fixed rho and variable phi

                For cylindrical cut
                    * `1`: axial cut with fixed phi and variable z
                    * `2`: circular cut with fixed z and variable phi"""

        self.polarization = 3
        """int: Polarization definition for the cut.
            GRASP Parameter ICOMP

            Values signify the following polarization definitions
                * `1`: Linear E_theta and E_phi.
                * `2`: Right hand and left hand circular (Erhc and Elhc).
                * `3`: Linear Eco and Ecx (Ludwig's third definition).
                * `4`: Linear along major and minor axes of the polarisation ellipse, Emaj and Emin.
                * `5`: XPD fields: E_theta/E_phi and E_phi/E_theta.
                * `6`: XPD fields: Erhc/Elhc and Elhc/Erhc.
                * `7`: XPD fields: Eco/Ecx and Ecx/Eco.
                * `8`: XPD fields: Emaj/Emin and Emin/Emaj.
                * `9`: Total power \\|E\\| and Erhc=Elhc."""

        self.field_components = 2
        """int: Number of field components
            GRASP Parameter NCOMP

            Value signifies:
                * `2`: for far field
                * `3`: for near field.  3rd component is always E_z"""

        self.data = numpy.ndarray((0, 0))
        """numpy.ndarray: Cut Data as complex array of field components.

        Shape of array is ``(v_num, field_components)``"""

    def read(self, lines):
        """Read cut from lines of text and parse as a cut, filing the
        parameters and data"""
        # Get the specification of the cut and parse it
        specline = lines.pop(0)
        specs = specline.split()

        # Make sure a stray comment line hasn't made it in at the start
        if specs[0] == "Field":
            if len(lines) != 0:
                specline = lines.pop(0)
                specs = specline.split()
            else:
                return

        self.v_ini = float(specs[0])
        self.v_inc = float(specs[1])
        self.v_num = int(specs[2])
        self.constant = float(specs[3])
        self.polarization = int(specs[4])
        self.icut = int(specs[5])
        self.field_components = int(specs[6])

        # Create the numpy structured array for the data
        self.data = numpy.zeros((self.v_num, self.field_components), dtype=complex)

        # Parse lines
        for i in range(self.v_num):
            lline = lines[i].split()
            if lline[0] == "Field":
                continue
            self.data[i, 0] = complex(float(lline[0]), float(lline[1]))
            self.data[i, 1] = complex(float(lline[2]), float(lline[3]))
            if self.field_components == 3:
                self.data[i, 2] = complex(float(lline[4]), float(lline[5]))

    @property
    def positions(self):
        """``numpy.array``: the positions of the data points in the cut file"""
        indices = numpy.arange(self.v_num, dtype=float)
        return self.v_ini + self.v_inc*indices

    def write(self):
        """Write local arrays to disk in GRASP cut file format"""
        pass

    def select_pos_range(self, pos_min, pos_max):
        """Returns a cut containing only the positions between pos_min
        and pos_max inclusive"""
        output = GraspSingleCut()

        output.text = self.text
        output.cut_type = self.cut_type

        output.v_ini = pos_min
        output.v_inc = self.v_inc
        output.v_num = math.ceil((pos_max - pos_min) / self.v_inc)
        output.constant = self.constant
        output.polarization = self.polarization
        output.icut = self.icut
        output.field_components = self.field_components

        # Find the indices of points just within the pos_min and pos_max
        # limits
        i_min = 0
        i_max = self.data.shape[0]
        for d in range(self.data.shape[0]):
            if self.positions[d] >= pos_min:
                if i_min == 0:
                    i_min = d
            if self.positions[d] >= pos_max:
                if i_max > d:
                    i_max = d

        # Set v_ini and v_num
        output.v_ini = self.positions[i_min]
        output.v_num = i_max - i_min + 1

        # Set data
        output.data = self.data[i_min:i_max, :]

        return output


class GraspCutSet:
    """Class for containing a set of cuts with a common parameter, such as
    frequency, beam, etc."""

    def __init__(self):
        self.cuts = []
        """list: List of :class:`.GraspCut` objects in the set."""


class GraspCut:
    """Class for reading, holding, manipulating and writing GRASP 9.3 format
    output cut files.

    A file may contain multiple sets of cuts (e.g. at different frequencies), with each set having several cuts with
    varying constant coordinate value (e.g. phi in a theta-phi grid).

    It's common for GRASP cut files to not contain information on the distinguishing parameter between cut sets,
    particular for cut file output from CHAMP calculations.  The user will have to supply this information, by e.g.
    creating a ``.frequencies`` list as an attributes to this object that stores the necessary information."""

    def __init__(self):
        self.cut_sets = []
        """list: List of :class:`.GraspCutSet`, each representing a set of cuts within the file"""

        self.cut_type = "spherical"
        """str: Describes the type of cut represented in the file.
        Options are:
            * ``spherical``: a cut defined on the surface of a sphere
            * ``planar``: a cut defined on a planar surface or the surface of a reflector
            * ``cylindrical``: a cut defined on the surface of a cylinder"""

        self.constants = []
        """list: A list of the constant values for each cut in each of the sets of cuts."""

    def read(self, fi):
        """Read the contents from filelike fi and parse into cut objects"""
        text = fi.readlines()

        self.constants = []
        # Constants contains a list of constants (typically phi angles)
        # When values are repeated, it indicates that a second set of cuts is
        # in the file

        self.cut_sets.append(GraspCutSet())
        cut_set = 0
        temp_text = []
        # Read through the file, splitting the lines into separate cuts
        for line in text:
            if len(line.split()) == 7:
                # We have the start of a new cut
                # Have we already collected a cut?
                if len(temp_text) > 2:
                    # Create new cut
                    new_cut = GraspSingleCut()
                    new_cut.read(temp_text)
                    if new_cut.constant in self.constants:
                        # We must start a new cut_set
                        self.cut_sets.append(GraspCutSet())
                        cut_set += 1
                        self.constants = []
                    self.cut_sets[cut_set].cuts.append(new_cut)
                    self.constants.append(new_cut.constant)
                    temp_text = []

            if len(line.strip()) > 0:
                temp_text.append(line)

        # Append the last cut to the file
        if temp_text:
            new_cut = GraspSingleCut()
            new_cut.read(temp_text)
            self.cut_sets[cut_set].cuts.append(new_cut)
            self.constants.append(new_cut.constant)

    def select_pos_range(self, pos_min, pos_max):
        """Return a new cut file object containing only the parts of
        each cut between pos_min and pos_max inclusive"""
        newcf = GraspCut()
        newcf.cut_type = "spherical"
        newcf.constants = self.constants
        for cut_set in self.cut_sets:
            newcs = GraspCutSet()
            for c in cut_set.cuts:
                newcs.cuts.append(c.select_pos_range(pos_min, pos_max))
            newcf.cut_sets.append(newcs)

        return newcf



import numpy as np
import numpy.ma as ma


def find_peak(field, comp=0, max_radius=None, min_radius=None):
    """Find the peak magnitude of a component in the field.

    Parameters:
        field ``GraspField``: The field to work on.
        comp int: The field component to look at.
        max_radius float: Ignore portions of the grid outside this radius from the center of the field.
        min_radius float: Ignore portions of the grid inside this radius from the center fo the field.

    Returns:
        x_peak float:, y_peak float: The x and y values of the peak value."""
    x_vals, y_vals = field.positions_1d

    f = abs(field.field[:, :, comp])
    if max_radius is not None:
        rad = field.radius_grid()
        rad_max_mask = ma.masked_greater(rad, max_radius)
        f = ma.array(f, mask=rad_max_mask.mask)

    if min_radius is not None:
        rad = field.radius_grid()
        rad_min_mask = ma.masked_less(rad, min_radius)
        f = ma.array(f, mask=rad_min_mask.mask)

    ny, nx = np.unravel_index(np.argmax(abs(f)), f.shape)
    x_peak = x_vals[nx]
    y_peak = y_vals[ny]

    return x_peak, y_peak


# find the center of illumination of the field
def find_center(field, comp=0, trunc_level=0.0, max_radius=None, min_radius=None):
    """Find the center of illumination by finding the "center of mass" of the field.

    Parameters:
        field ``GraspField``: The field to work on.
        comp int: The field component to look at.
        trunc_level float: Ignore the contributions from portions of the grid below this field level.
        max_radius float: Ignore portions of the grid outside this radius from the center of the field.
        min_radius float: Ignore portions of the grid inside this radius from the center fo the field.

    Returns:
        x_cent float, y_cent float: The x and y values of the center of the field."""
    xv, yv = field.positions

    f = abs(field.field[:, :, comp])
    if trunc_level != 0.0:
        f = ma.masked_less_equal(f, trunc_level)
        xv = ma.array(xv, mask=f.mask)
        yv = ma.array(yv, mask=f.mask)

    if max_radius is not None:
        rad = field.radius_grid()
        rad_max_mask = ma.masked_greater(rad, max_radius)
        f = ma.array(f, mask=rad_max_mask.mask)
        xv = ma.array(xv, mask=rad_max_mask.mask)
        yv = ma.array(yv, mask=rad_max_mask.mask)

    if min_radius is not None:
        rad = field.radius_grid()
        rad_min_mask = ma.masked_less(rad, min_radius)
        f = ma.array(f, mask=rad_min_mask.mask)
        xv = ma.array(xv, mask=rad_min_mask.mask)
        yv = ma.array(yv, mask=rad_min_mask.mask)

    x_illum = xv * f
    y_illum = yv * f

    norm = np.sum(f)

    x_cent = np.sum(x_illum) / norm
    y_cent = np.sum(y_illum) / norm

    return x_cent, y_cent



# test_cut.py

import pytest
from graspfile import cut
from pytest import approx

test_cut_file = "tests/test_data/grasp_files/cut_files/example_GRASP_10-0-1_spherical_polar_linear_farfield.cut"
"""TICRA Tools 19.1 GRASP Cut file"""


@pytest.fixture
def empty_grasp_cutfile():
    """Return an empty  instance."""
    return cut.GraspCut()


@pytest.fixture
def cut_file():
    """Return a file object that points to a GRASP Cut file"""
    return open(test_cut_file)


@pytest.fixture
def filled_grasp_cut_file(empty_grasp_cutfile, cut_file):
    """Return a GraspCutFile instance filled from the cut_file."""
    empty_grasp_cutfile.read(cut_file)
    cut_file.close()
    return empty_grasp_cutfile


@pytest.fixture
def filled_grasp_cut(filled_grasp_cut_file):
    """Return a GraspField instance from the filled_grasp_grid fixture"""
    return filled_grasp_cut_file.cut_sets[0].cuts[0]


def test_loading_cut_file(filled_grasp_cut_file):
    """Test loading cut from a TICRA Cut file."""
    # check that enough frequencies and fields were read
    assert len(filled_grasp_cut_file.cut_sets[0].cuts) > 0

    # Check that parameters were read correctly
    assert filled_grasp_cut_file.cut_type in ["spherical", "planar", "cylindrical"]
    assert len(filled_grasp_cut_file.constants) > 0


def test_select_pos_range_file(filled_grasp_cut_file):
    """Check that returning a sub range works"""
    filled_grasp_cut = filled_grasp_cut_file.cut_sets[0].cuts[0]

    v_min = filled_grasp_cut.v_ini
    v_max = filled_grasp_cut.v_ini + \
        (filled_grasp_cut.v_num - 1) * filled_grasp_cut.v_inc
    v_cent = (v_min + v_max) / 2.0
    v_range = (v_max - v_min) / 2

    pos_max = v_cent + v_range / 2
    pos_min = v_cent - v_range / 2

    newcf = filled_grasp_cut_file.select_pos_range(pos_min, pos_max)

    assert len(newcf.cut_sets) == len(filled_grasp_cut_file.cut_sets)
    assert len(newcf.cut_sets[0].cuts) == len(filled_grasp_cut_file.cut_sets[0].cuts)


def test_loading_cut(filled_grasp_cut):
    """Test the individual field loaded as part of filled_grasp_grid"""
    # check that cut parameters were filled correctly

    assert type(filled_grasp_cut.v_ini) is float
    assert type(filled_grasp_cut.v_inc) is float
    assert type(filled_grasp_cut.v_num) is int

    assert filled_grasp_cut.cut_type in range(1, 4)
    assert type(filled_grasp_cut.constant) is float
    assert filled_grasp_cut.icut in range(1, 3)
    assert filled_grasp_cut.polarization in range(1, 10)
    assert filled_grasp_cut.field_components in [2, 3]

    # Check that the shape of the field is consistent with grid parameters
    data_shape = filled_grasp_cut.data.shape

    assert data_shape[0] == filled_grasp_cut.v_num
    assert data_shape[1] == filled_grasp_cut.field_components
